=head1 NAME

Eidetic::Racks

=head1 SYNOPSIS

A viewer for computer equipment racks, that pulls its data from a
database using Eidetic middleware.

=head1 DESCRIPTION

This perl module can be used to generate a display of a set of computer racks.
It generates this in the form of HTML tables, with autogenerated graphics (via GD)

=head1 FUNCTIONS

=cut

package Eidetic::Racks;

use strict;
use Carp   qw(confess);
use Eidetic::DBObject;
use File::Spec::Functions qw(:ALL);
use GD;

use Exporter;
use DynaLoader;

use vars qw( @ISA %FIELDS $VERSION);
use constant DEBUGGING => 0;
our @ISA = qw( Eidetic::DBObject Exporter DynaLoader );
our @EXPORT = qw( hostnameToId lookForIds );

my $NAME = __PACKAGE__;
my $TABLE = "host";
my $KEY   = "uid";
my $NAME_FIELD = "descriptor";
my $POS_FIELD = "U_count_from_bottom";
my $SIZE_FIELD = "Size_in_rack_u";
my $RACK_FIELD = "Rack_number";
my $HEIGHT_U_FIELD = "Height_U";
my $RACK_FORMAT = "%03d";

$VERSION = 0.09;

use fields qw(
              _hosts
              _racks
              _database
              _filename
              _frame
              _rack_ids
              _rack_inches
              _max_u
              _u_pixels
              _rack_width
              _cell_border_width
              _rack_border_width
              _img_dir
              _img_url
              _overwrite_pngs
              _host_viewer
              );

=head2 hostnameToId(DB, HOSTNAME)

Return the ID for the given hostname, or undef if not found.

=cut

sub hostnameToId
{
    my $db = shift || die "No db specified\n";
    my $hostname = shift || die "No hostname specified\n";
    
    confess "No database connection" unless (defined($db));

    my $query = "SELECT $KEY FROM $TABLE WHERE descriptor = ?";

    my $ref=$db->hashedQuery($query,$KEY,($hostname));

    # did something blow up?
    return ()
        unless (defined($ref));

    return sort keys %{ $ref };
}


=head2 lookForUIDs(DB, WHERE/FIELDS)

Returns a list of match UIDs from the Eidetic::Database, using
field/value pairs passed in the FIELDS hash, or single WHERE clause.

=cut

sub lookForIds
{
    my $db = shift;
    my %query;
    my $query;
    my @values;

    confess "No database connection" unless (defined($db));

    if ((scalar(@_) % 2)==1) {
        # odd arg means direct "WHERE" statement
        $query="SELECT $KEY FROM $TABLE WHERE ".shift;
        @values=@_;
    } else {
        %query=@_;

        my @fields;
        foreach my $field (keys %query)
        {
            push(@fields,"$field = ?");
            push(@values,$query{$field});
        }
        if (@fields) {
            $query="SELECT $KEY FROM $TABLE WHERE ".join(" AND ",@fields);
        }
        else {
            $query="SELECT $KEY FROM $TABLE";
        }
    }
    
    my $ref=$db->hashedQuery($query,$KEY,@values);

    # did something blow up?
    return ()
        unless (defined($ref));

    return sort keys %{ $ref };
}

=head2 draw_png 

General purpose GD-based drawing routine

=cut

sub draw_png {
    my %args = @_;
    my $img_path   = $args{'img_path'};
    my $img_width  = $args{'img_width'};
    my $img_height = $args{'img_height'};
    my $content    = $args{'content'};
    my $framed     = $args{'framed'} ? 1 : 0;

    # Generate the img for the host's name
    my $im = new GD::Image($img_width,$img_height);

    # Allocate some colors
    # First color allocated is the background
    my $bg;
    my $white;
    if (defined($args{'bg_color'}) && 
        $args{'bg_color'} =~ /^#?(..)(..)(..)$/) {
        $bg = $im->colorAllocate(hex($1),hex($2),hex($3));
        if (hex($1) != 255 ||
            hex($2) != 255 ||
            hex($3)) {
            $white = $im->colorAllocate(255,255,255);
        } else {
            # Background IS white already...
            $white = $bg;
        }
    } else {
        $bg = $im->colorAllocate(255,255,255);
        $white = $bg;
    }
    my $black = $im->colorAllocate(0,0,0);
    my $red   = $im->colorAllocate(255,0,0);
    my $green = $im->colorAllocate(0,255,0);
    my $blue  = $im->colorAllocate(0,0,255);
                
    # Make the background transparent
    $im->transparent($bg) if ($white == $bg);

    # Put a black frame around the picture
    if ($framed) {
      $im->rectangle(0,0,$img_width-1,$img_height-1,$black);
    }
            
    # Write on it
    if (defined($content)) {
      $im->string(gdSmallFont,2,0,$content,$black);
    }
            
    # Convert the image to PNG and $text .= it on standard output
    open (IMGFILE, ">$img_path") || die "Could not open $img_path for writing: $!";
    # Make sure we are writing to a binary stream
    binmode IMGFILE;
            
    print IMGFILE $im->png;
    close (IMGFILE) || die "Could not close $img_path: $!";
}

=head2 draw_host_png

Creates a png image for given hosts info, to the given height and
width, stored to the file system at the specified path (dir/filename)

=cut

sub draw_host_png {
    my %args = @_;
    my $host = $args{'host'};
    $args{'content'} = $host->{$NAME_FIELD};

    draw_png(%args);
}

=head2 new(DB, FIELDS)

Builds an object for a set of racks containing computer machines
(hosts).

database            Eidetic::Database object (optional)
filename            File to read host data from (unimplemented)
frame               Frame of racks to use (required unless rack_ids defined)
rack_ids            ID's of racks (required unless frame defined)
max_u               Height of the racks in U's (default 42)
u_pixels            Number of pixels per U for drawing (default 12)
rack_width          Width in pixels to make a rack (default 144)
cell_border_width   Width for cell borders (default 0)
rack_border_width   Width for rack borders (default 1)
img_dir             Directory to write host imgs to (default current dir)
img_url             URL for accessing host imgs (default same as the emitted html)
overwrite_pngs      If true, forces pngs to be regenerated

=cut

sub new {
    my ($this, %args) = @_;
    my $class = ref($this) || $this;
    my $self = bless [\%FIELDS], $class;
    
    # Bring in the args
    while (my ($field, $value) = each %args) {
        confess "Invalid parameter to $NAME '$field'"
            unless (exists $FIELDS{"_$field"});
        $self->{"_$field"} = $value;
    }

    # Required parameters
    if (!defined($self->{_frame}) &&
        !defined($self->{_rack_ids})) {
        confess "$NAME requires either 'frame' or 'rack_ids' parameter"
    }

    # Specify defaults
    $self->{_u_pixels}            ||= 12;
    $self->{_rack_inches}         ||= 19;
    # To calculate pixel width, assume _rack_inches plus 1 inch on
    # either side.  And then divide that by 1.75 to get U width, and
    # multiply by u_pixels.
    $self->{_rack_width}          ||= (($self->{_rack_inches} + 2) / 1.75) *
                                        $self->{_u_pixels};
    $self->{_cell_border_width}   ||= 0;
    $self->{_rack_border_width}   ||= 1;
    $self->{_img_dir}             ||= '.';
    $self->{_img_url}             ||= '';
    $self->{_overwrite_pngs}      ||= 0;
    $self->{_host_viewer}         ||= '';

    # Assert existance of required directories & files
    confess "Image output directory $self->{_img_dir} does not exist"
        unless (-d $self->{_img_dir});

    # Tack on a trailing / to the image url if not already given
    if ($self->{_img_url} ne '' && $self->{_img_url} !~ /\/$/) {
        $self->{_img_url} .= '/';
    }
    
    # Get host records
    if ($self->{_database}) {
        my $DB = $self->{_database};
        # Get the data for the hosts in this rack from the database
        confess "No database connection" unless (defined($DB));

        # Get ready for hot database-on-database action!
        my ($filter,$sort,$sth);

        # Find out the list of racks
        if (defined($self->{_frame})) {
            $filter = "frame_to_rack.frame_uid = ".$self->{_frame};
            $sort   = "frame_to_rack.rack_uid";

            $sth = $DB->selectRecords('frame_to_rack', $filter, $sort);
            # wipe out the rack id list
            undef @{$self->{_rack_ids}};
            while (my $row = $sth->fetchrow_hashref) {
                push(@{$self->{_rack_ids}},$row->{'rack_uid'});
            }
        }

        # FIXME this is specific to OSDL's rack number formatting...
        foreach my $rack (@{$self->{_rack_ids}}) {
            $rack=sprintf($RACK_FORMAT,$rack);

            $filter = "rack.uid = $rack";
            $sort = "rack.uid";
            $sth = $DB->selectRecords('rack_type<-rack',$filter,$sort);

            while (my $row = $sth->fetchrow_hashref) {
                $self->{_racks}->{$rack} = $row;
            }
        }

        $filter = "host.$RACK_FIELD IN (".join(',',@{$self->{_rack_ids}}).")";
        warn "Filter is $filter\n" if (DEBUGGING);

        $sort   = "host.$RACK_FIELD";
        $sth = $DB->selectRecords('host_type<-host', $filter, $sort);
        # TODO:  Sort hosts by rack number & u_count_from_bottom, ascending

        die "No hosts found" unless ($sth != 0);

        warn ("Number of records:  " . $sth->rows . "\n") if (DEBUGGING);

        while (my $row = $sth->fetchrow_hashref) {
            $row->{$POS_FIELD} ||= 0;
            $row->{$SIZE_FIELD} ||= 1;
            if (DEBUGGING>1) {
                    foreach my $key (keys %$row) {
                        next unless ($row->{$key});
                        warn "$key = " . $row->{$key} . "\n";
                    }
            }
            $self->{_hosts}->{$row->{uid}} = $row;
            warn "Row:  $row->{uid} $row->{descriptor}\n" if (DEBUGGING);
        }

        warn ("Number of hosts:  " . scalar(keys %{$self->{_hosts}}) . "\n") if (DEBUGGING);

    } elsif ($self->{_filename}) {
        # Load hosts from filename
        carp("Host config file '$self->{_filename}' could not be found\n")
        unless (-e $self->{_filename});

        $self->{_hosts} = do "$self->{_filename}";
    } else {
        warn "No host input database or filename provided.\n";

    }

    my $max=0;
    foreach my $rack (sort keys %{$self->{_racks}}) {
        $max = $self->{_racks}->{$rack}->{$HEIGHT_U_FIELD}
            if ($self->{_racks}->{$rack}->{$HEIGHT_U_FIELD}>$max);
    }
    $self->{_max_u} = $max == 0 ? 42 : $max;

    return $self;
}


=head2 draw_u_ruler(U_COUNT)

Creates an HTML table measured out in 1 U increments

=cut 

sub draw_u_ruler {
    my ($self, $u_count, $height, $width) = @_;

    return '' unless ($u_count && $u_count>0);

    my $text .= qq(    <!-- U mark ruler -->\n);
    $text = qq(    <TABLE BORDER="0" BGCOLOR="white" CELLSPACING="$self->{_cell_border_width}" CELLPADDING="0");
    for (my $i=$u_count; $i--; ) {
        my $blank_path = catfile($self->{_img_dir}, $i.'umark.png');
        my $blank_url  = $self->{_img_url} . $i.'umark.png';
        my $blank_width = 18;
        my $blank_height = 1 * $self->{_u_pixels};
        if (! -e $blank_path) {
            draw_png('img_path'   => $blank_path,
                     'img_width'  => $blank_width,
                     'img_height' => $blank_height,
                     'bg_color'   => "#FFFFFF",
                     'content'    => "$i",
                     'framed'     => 1);
        }

        $text .= qq(     <TR>\n);
        $text .= qq(      <TD BGCOLOR="#666666">\n);
        $text .= qq(       <IMG SRC="$blank_url" HEIGHT="$blank_height" WIDTH="$blank_width" ALT="">\n);
        $text .= qq(      </TD>\n);
        $text .= qq(     </TR>\n);
    }
    $text .= qq(</TABLE>);
    return $text;
}

=head2 getHeat($rack)

Get the total BTU and cooling info for a rack.
Returns array of ($btu,$cooling_tons)

=cut

sub getHeat
{
    my ($self,$rack) = @_;

    # Track heat per rack
    my $btus=0;
    my $cooling=0;
    
    my @hosts = sort { 
                $a->{$POS_FIELD} <=> $b->{$POS_FIELD} 
    } grep($_->{$RACK_FIELD} == $rack, values %{$self->{_hosts}});

    foreach my $host (@hosts) {
        $btus   +=defined($host->{Max_btus}) ? $host->{Max_btus} : 0;
        $cooling+=defined($host->{Cooling_tons_max}) ? 
                    $host->{Cooling_tons_max} : 0;
    }

    return ($btus,$cooling);
}

=head2 sanityCheck($rack)

Produces text lines of warnings about a given Rack.

=cut

sub sanityCheck
{
    my ($self,$rack) = @_;

    # An array to gather all rack warning strings
    my @warning;

    # Track rack U reservation
    my @u_reservation;

    my @hosts = sort { 
                $a->{$POS_FIELD} <=> $b->{$POS_FIELD} 
    } grep($_->{$RACK_FIELD} == $rack, values %{$self->{_hosts}});

    foreach my $host (@hosts) {
            # Clean up brand warnings
            my $name = $host->{descriptor};
            my $brand = defined($host->{Brand}) ? $host->{Brand} : "no-brand";
            my $brand_model_name = defined($host->{Brand_model_name}) ?
                                    $host->{Brand_model_name} :
                                    "no-brand-modelname";

            # Verify at least brand and model names
            if (!defined($host->{Brand})) {
                push(@warning,
                     "$brand $brand_model_name (host $name) needs Brand name");
            }
            if (!defined($host->{Brand_model_name})) {
                push(@warning,
                     "$brand $brand_model_name (host $name) ".
                     "needs Brand_model_name");
            }
            
            # Verify host type if loading from database
            if ($self->{_database} && !defined($host->{Host_type_uid})) {
                push(@warning,
                     "$brand $brand_model_name (host $name) needs host_type");
            }

            # Verify U position
            if (!defined($host->{$POS_FIELD})) {
                push(@warning, "Host $name missing U position");
                $host->{$POS_FIELD}=0;
            }

            # Verify U size
                if ($host->{$SIZE_FIELD} < 1) {
                push(@warning, "Host $name missing U size");
                    $host->{$SIZE_FIELD} = 1;
                }

            # Reserve the U slots
            my $loc;
            for ($loc=0;$loc<$host->{$SIZE_FIELD};$loc++) {
                push(@{ $u_reservation[$host->{$POS_FIELD}+$loc] }, $host);
            }

            # Verify BTU information
            if (!defined($host->{Max_btus})) {
                push(@warning,
                     "$brand $brand_model_name (host $name) needs BTU info");
            }
            # Leave this here until we find a device that really doesn't
            # have any BTU
            else {
                if ($host->{Max_btus}==0) {
                    push(@warning,
                         "$brand $brand_model_name (host $name) ".
                         "has Max_btus == 0 ?!");
                }
            }

            # Verify cooling information
            if (!defined($host->{Cooling_tons_max})) {
                push(@warning,
                     "$brand $brand_model_name (host $name) needs cooling info");
            }
            # Leave this here until we find a device that really doesn't
            # have any cooling needs info
            else {
                if ($host->{Cooling_tons_max}==0) {
                    push(@warning,
                         "$brand $brand_model_name (host $name) ".
                         "has Cooling_tons_max == 0 ?!");
                }
            }
    }

    # Verify U position reservations
    foreach my $loc (0 .. $#u_reservation) {
        if (exists($u_reservation[$loc])) {
            my @list = @{ $u_reservation[$loc] };

            #$text .= "<!-- rack $rack u $loc: ".
            #         join(" ",map { $_->{descriptor} } @list).
            #         " -->\n";

            if ($#list > 0) {
                push(@warning,"U position $loc in conflict: ".
                     join(" ",map { $_->{descriptor} } @list));
            }
        }
    }


    return @warning;
}

=head2 toHtml

Renders a view of racks in HTML and PNG, suitable for plugging into a
webpage.

Arguments:
    rack_heat  => 1    Report each racks's heat information
    frame_heat => 1    Report entire frame's heat information
    heat       => 1    Report all heat info
    warn       => 1    Report any information warnings
    alpha      => 1    Report hosts in alphabetical order with U locations
    umarks     => 1    Shows U locations on both sides

=cut

sub toHtml {
    my $self = shift;
    my %args = @_;
    my $text = '';

    # Set up argument defaults
    $args{'rack_heat'}  ||= 0;
    $args{'frame_heat'} ||= 0;
    # heat should be undef by default.  0 means "turn off all heat repots"
    #                                   1 means "turn on all heat reports"
    #$args{'heat'}       ||= 0;
    $args{'warn'}       ||= 0;
    $args{'alpha'}      ||= 0;
    $args{'umarks'}     ||= 0;
    $args{'javascript_images'} ||= 0;

    # Heat summaries
    my $sum_btus=0;
    my $sum_cooling=0;

    return 'No rack ids specified' unless (@{$self->{_rack_ids}});

    # A table to hold all the racks together
    $text .= qq(<TABLE BORDER="0" CELLSPACING="1" CELLPADDING="1">\n);

    # Report Rack numbers at the top
    $text .= qq( <TR>\n);

    if ($args{'umarks'}) {
        $text .= qq(  <TD>&nbsp;</TD>\n);
    }
    foreach my $rack (@{$self->{_rack_ids}}) {
        $text .= qq(  <TD>\n);
        $text .= qq(   <CENTER><B>$rack</B></CENTER>\n);
        $text .= qq(  </TD>\n);
    }
    if ($args{'umarks'}) {
        $text .= qq(  <TD>&nbsp;</TD>\n);
    }
    $text .= qq( </TR>\n);
        
    # Figure out 1U image
    my $blank_path = catfile($self->{_img_dir}, '1u.png');
    my $blank_url  = $self->{_img_url} . '1u.png';
    my $blank_width = $self->{_rack_width};
    my $blank_height = 1 * $self->{_u_pixels};
    if (! -e $blank_path) {
        draw_host_png('img_path'   => $blank_path,
                      'img_width'  => $blank_width,
                      'img_height' => $blank_height,
                      'bg_color'   => "#999999",
                      'framed'     => 1);
    }

    # a stand of racks, starting from top, one U at a time
    $text .= qq( <TR BGCOLOR="black">\n);
    if ($args{'umarks'}) {
        # Print the U counts up the left side
        $text .= qq(  <TD BORDER="$self->{_rack_border_width}" VALIGN="bottom" BGCOLOR="black">\n\n);
        $text .= $self->draw_u_ruler($self->{_max_u});
        $text .= qq(  </TD>\n);
    }

    foreach my $rack (@{$self->{_rack_ids}}) {
        # Since racks may have different height, align them to the bottom
        $text .= qq(  <TD BORDER="$self->{_rack_border_width}" VALIGN="bottom" BGCOLOR="black">\n\n);

        # a table to hold the rack itself
        $text .= qq(   <!-- Rack $rack -->\n);
        $text .= qq(   <TABLE BORDER="0" CELLPADDING="0" CELLSPACING="$self->{_cell_border_width}">\n);

        my @rows = ();
        my $current_u = 0;

        # Sort hosts by U position
        my @hosts = sort { 
                $a->{$POS_FIELD} <=> $b->{$POS_FIELD} 
            } grep($_->{$RACK_FIELD} == $rack, values %{$self->{_hosts}});
        
        foreach my $host (@hosts) {
            warn ("UID:  " . $host->{uid} . "\n") if DEBUGGING;
            warn ("Keys:  " . join(',',keys(%{$host})) . "\n") if DEBUGGING;
            warn ("Rack number : " . $host->{$RACK_FIELD} . "\n") if DEBUGGING;
            #next unless (exists $host->{$RACK_FIELD} && $host->{$RACK_FIELD} == $rack);

            # Fill in any needed blank spaces
            #
            while ($host->{$POS_FIELD} > $current_u) {
                my $row = qq(    <TR BGCOLOR="silver">\n);
                $row .=   qq(      <TD><IMG SRC="$blank_url" WIDTH="$blank_width" HEIGHT="$blank_height" BORDER="0"></TD>\n);
                $row .=   qq(    </TR>\n);
                push @rows, $row;
                $current_u ++;
            }

            my $img_path = catfile($self->{_img_dir}, $host->{descriptor} . '.png');
            my $img_url  = $self->{_img_url} . $host->{$NAME_FIELD} . '.png';
            my $img_width = $self->{_rack_width};
            my $img_height = $host->{$SIZE_FIELD} * $self->{_u_pixels};

            # stretch the image to deal with the borders on non-aligned hosts
            $img_height += ($host->{$SIZE_FIELD}-1) * ($self->{_cell_border_width});
            if ($self->{_overwrite_pngs} || ! -e $img_path) {
                draw_host_png('img_path'   => $img_path,
                              'img_width'  => $img_width,
                              'img_height' => $img_height,
                              'host'       => $host,
                              'framed'     => 1);
            }
            # Display the host
            my $host_url = $self->{_host_viewer};
            $host_url =~ s/\[\%\s*uid\s*\%\]/$host->{uid}/g;
            $host_url =~ s/\[\%\s*descriptor\s*\%\]/$host->{descriptor}/g;
            my $row = qq(    <TR BGCOLOR="white">\n);
            $row .=   qq(     <TD>);
            if ($host_url) {
                $row .=   qq(<A HREF="$host_url">);
            }
            if ($args{'javascript_images'}) {
                $row .=   qq(<IMG NAME="$host->{descriptor}" SRC="$img_url" WIDTH="$img_width" HEIGHT="$img_height" BORDER="0" onMouseOver="$host->{descriptor}_MouseOver()" onMouseOut="$host->{descriptor}_MouseOut()">);
            } else {
                $row .=   qq(<IMG SRC="$img_url" WIDTH="$img_width" HEIGHT="$img_height" BORDER="0">);
            }
            if ($host_url) {
                $row .=   qq(</A>);
            }
            $row .=   qq(</TD>\n);
            $row .=   qq(     </TD>\n);
            $row .=   qq(    </TR>\n);
            push @rows, $row;
            $current_u = $host->{$POS_FIELD} + $host->{$SIZE_FIELD};

        }
        
        # Fill in remaining blank spaces
        while ($current_u < $self->{_max_u}) {
            my $row = qq(    <TR BGCOLOR="silver">\n);
            if (defined($self->{_racks}->{$rack}) &&
                $current_u >= $self->{_racks}->{$rack}->{$HEIGHT_U_FIELD}) {
                # non-existant rack position
                $row .=   qq(      <TD></TD>\n);
            }
            else {
                # empty rack position
                $row .=   qq(      <TD><IMG SRC="$blank_url" WIDTH="$blank_width" HEIGHT="$blank_height" BORDER="0"></TD>\n);
            }
            $row .=   qq(    </TR>\n);
            push @rows, $row;
            $current_u ++;
        }

        # Reverse rows 
        foreach my $row ( reverse @rows ) {
            $text .= $row;
        }

        # Close table
        $text .= qq(   </TABLE>\n);
        $text .= qq(  </TD>\n);
    }

    $text .= qq(  </TD>\n) unless (! @{$self->{_rack_ids}});
    # Rack U-count ruler
    if ($args{'umarks'}) {
        $text .= qq(  <TD BORDER="$self->{_rack_border_width}" VALIGN="bottom" BGCOLOR="black">\n\n);
        $text .= $self->draw_u_ruler($self->{_max_u});
        $text .= qq(  </TD>\n);
    }                          
    $text .= qq( </TR>\n);

    # Report Rack numbers at the bottom
    $text .= qq( <TR>\n);
    if ($args{'umarks'}) {
        $text .= qq(  <TD>&nbsp;</TD>\n); # umark spacing
    }
    foreach my $rack (@{$self->{_rack_ids}}) {
        $text .= qq(  <TD>\n);
        $text .= qq(   <CENTER><B>$rack</B></CENTER><BR>\n);

        # Calculate heat
        my ($btus,$cooling) = $self->getHeat($rack);
        $sum_btus+=$btus;
        $sum_cooling+=$cooling;

        # Report rack heat, if asked
        if ((defined($args{'heat'}) && $args{'heat'}) ||
            (!defined($args{'heat'}) && $args{'rack_heat'})
            ) {
            $text .= qq($btus BTU max<BR>\n);
            $text .= sprintf("%.1f tons cooling max<BR>\n", $cooling);
        }
        $text .= qq(  </TD>\n);
    }
    if ($args{'umarks'}) {
        $text .= qq(  <TD>&nbsp;</TD>\n); # umark spacing
    }
    $text .= qq( </TR>\n);
    $text .= qq(</TABLE>\n\n);

    $text .= qq(<P>\n);

    # Report extra text using a seprate table so we don't blow up the
    # graphic formatting
    $text .= qq(<TABLE BORDER="0" CELLSPACING="1" CELLPADDING="1">\n);
    if ((defined($args{'heat'}) && $args{'heat'}) ||
        (!defined($args{'heat'}) && $args{'frame_heat'})
        ) {
        $text .= qq( <TR>\n);
        $text .= qq(  <TD>\n);
        $text .= qq(Frame BTU max: $sum_btus<BR>\n);
        $text .= sprintf("Frame tons cooling max: %.1f<BR>\n", $sum_cooling);
        $text .= qq(  </TD>\n);
        $text .= qq( </TR>\n);
    }
    if ($args{'warn'}) {
        $text .= qq( <TR>\n);
        $text .= qq(  <TD>\n);
        $text .= qq(   <FONT COLOR="red">\n);
        foreach my $rack (@{$self->{_rack_ids}}) {
            foreach my $line ($self->sanityCheck($rack)) {
                $text .= qq(Rack $rack: $line<BR>\n);
            }
        }
        $text .= qq(   </FONT>\n);
        $text .= qq(  </TD>\n);
        $text .= qq( </TR>\n);
    }
    $text .= qq(</TABLE>\n\n);

    if ($args{'alpha'}) {
        $text .= qq(<TABLE BORDER="0">\n);
        $text .= qq( <TR>\n);
        $text .= qq(  <TH>Host</TH>\n);
        $text .= qq(  <TH>Rack</TH>\n);
        $text .= qq(  <TH>&nbsp;</TH>\n);
        $text .= qq(  <TH>U</TH>\n);
        $text .= qq( </TR>\n);
        foreach my $host (sort {
                            $a->{$NAME_FIELD} cmp $b->{$NAME_FIELD}
                               } values %{$self->{_hosts}}) {
            my $rack=sprintf($RACK_FORMAT,$host->{$RACK_FIELD});
            $text .= qq( <TR>\n);
            $text .= qq(  <TD ALIGN="left">$host->{$NAME_FIELD}</TD>\n);
            $text .= qq(  <TD ALIGN="right">$rack</TD>\n);
            $text .= qq(  <TD>&nbsp;</TD>\n);
            $text .= qq(  <TD ALIGN="right">$host->{$POS_FIELD}</TD>\n);
            $text .= qq( </TR>\n);
        }
        $text .= qq(</TABLE>\n\n);
    }

    return $text;
}

1;

# vi:set ai ts=4 sw=4 expandtab:
